import functools
from itertools import chain
import typing as t

from mb.messages import Message

P = t.ParamSpec("P")
R = t.TypeVar("R")
M = t.TypeVar("M", bound=Message | None)


class TrackedHandler(t.Protocol[P, R, M]):
    calls: list[M]
    __call__: t.Callable[P, R]


def tracked_handler(func: t.Callable[P, R]) -> TrackedHandler[P, R, M]:
    """
    Tracks the calls to a handler in the :attr:`calls`.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> R:
        for argument in chain(args, kwargs.values()):
            if isinstance(argument, Message):
                wrapper.calls.append(argument)  # type:ignore
                break
        else:
            wrapper.calls.append(None)  # type:ignore
        return func(*args, **kwargs)

    wrapper.calls = []  # type:ignore
    return wrapper  # type:ignore



import typing as t

if t.TYPE_CHECKING:
    from mb.commands import Command

# TODO: Sufix all errors with Error (and remove some prefixes?)
# TODO: Move some message str to the __init__ of these exceptions

# --------------------------------------
# Base exceptions
# --------------------------------------


class CQBaseError(BaseException):
    """
    Uncatchable error
    """


class CQProgrammingError(CQBaseError, RuntimeError):
    """
    Uncatchable error. Code violates some invariant
    """


class CQError(Exception):
    """
    Library root exception
    """


class ConfigError(CQError):
    """
    General configuration error
    """


# --------------------------------------
# Message definitions
# --------------------------------------


class InvalidMessageName(ConfigError, TypeError):
    """
    The Message.NAME must be an string
    """


class DuplicatedMessageName(ConfigError, TypeError):
    """
    Two messages have the same NAME
    """


# --------------------------------------
# Handler configuration
# --------------------------------------


class DuplicatedCommandHandler(ConfigError):
    """
    A Command can have only one handler
    """


class MissingCommandHandler(ConfigError, RuntimeError):
    """
    No handler found for a Command
    """

    def __init__(self, command: "Command"):
        self.message = f"Missing handler for command: '{command}'"
        self.command = command
        super().__init__(self.message, self.command)


class InvalidMessageType(ConfigError, TypeError):
    """
    The Bus only handles Messages
    """


# --------------------------------------
# Handling messages
# --------------------------------------


class InvalidMessage(CQError, TypeError):
    """
    The Bus only handles Messages
    """


class InjectError(CQError, RuntimeError):
    """
    The argument couldn't be injected
    """

    def __init__(self, argument: str, handler: t.Callable):
        message = (
            "Cannot inject the {argument} argument in handler '{handler}'. "
            "No candidates found by name or annotation."
        )
        super().__init__(message, argument, handler)
        self.message = message
        self.argument = argument
        self.handler = handler


# --------------------------------------
# Databases
# --------------------------------------


class InvalidTransactionState(CQError):
    """
    Likely an internal error
    """


# --------------------------------------
# Unit of Work
# --------------------------------------


class UowContextRequired(CQError):
    """
    Code is using the UnitOfWork outside the context or
    the begin/commit/rollback calls are unpaired.

        >>> uow = UnitOfWork(bus)
        >>> with uow:
        ...     uow.emit_event(event)  # Good
        ... uow.emit_event(event)  # Bad
    """



import typing as t
from contextlib import contextmanager

from django.db import DEFAULT_DB_ALIAS
from django.db.transaction import atomic
from django.db.transaction import get_connection

from mb.unit_of_work.nested import NestedUnitOfWork
from mb.unit_of_work.utils.events_collector import DedupeEventsFifo

if t.TYPE_CHECKING:
    from mb.bus import MessageBus
    from mb.unit_of_work.utils.events_collector import EventsCollector

Using = t.Union[list[str], str, None]


class DjangoUnitOfWork(NestedUnitOfWork):
    """
    Integrates the UoW with the Django ORM machinery.

    --------------
    Models binding
    --------------

    Binds this UoW and MessageBus to the connections of the current thread/context,
    providing a way for Model instances to access the bus by themselves when running
    within an UoW transaction.

        >>> from mb.contrib.django.models import BusModel
        >>> class User(BusModel):
        ...    def save(self, *a, **kw):
        ...        super().save(*a, **kw)
        ...        self.uow.emit_event(UserCreated(self.id))
        ...
        >>> # GOOD
        >>> uow = DjangoUnitOfWork(bus)
        >>> with uow.atomic():
        ...     user = User(...)
        ...     user.save()  # Emits UserCreated in the current uow.
        ...
        >>> # BAD
        >>> user = User(...)
        >>> user.save()  # Raises UowContextRequired!

    ---------------------------------
    UoW transactions VS Django atomic
    ---------------------------------

    Note that the events are handled when the outermost UoW transaction is closed,
    regardless of the database transaction state. Therefore, it's highly recommended
    that your outermost Bus TX wraps the outermost Database TX, otherwise your events
    will be handled before the performing the Database Commit.

    BAD:

        >>> uow = DjangoUnitOfWork(bus)
        >>> with atomic():
        ...     with uow:
        ...         user.create()
        ...         uow.emit_event(UserCreated(user.id))  # queued
        ...     # UoW/Bus handles the events here (e.g., sending an email)
        ...     pass
        ... # Django commits but the database raises an error!!
        ... # You have sent an email to a user that was never created (⩾﹏⩽)
        ... pass

    GOOD:

        >>> uow = DjangoUnitOfWork(bus)
        >>> with uow:
        ...     with atomic():
        ...         user.save()
        ...         uow.emit_event(UserCreated(user.id))  # queued
        ... # Django commits the database and if there aren't errors
        ... # the UoW/Bus handles the events just before
        ... pass

    GOOD (shorter):

        >>> uow = DjangoUnitOfWork(bus)
        >>> with uow.atomic():
        ...     user.save()
        ...     uow.emit_event(UserCreated(user.id))  # queued

    -------------------------------------------------
    Savepoint rollbacks and other undesired scenarios
    -------------------------------------------------

    The following scenario is not supported yet:

        >>> uow = DjangoUnitOfWork(bus)
        >>> with uow.atomic():
        >>>     connection = get_connection()
        >>>     savepoint_id = connection.savepoint_ids[-1]
        >>>     # UoW collects events of the following function
        >>>     do_stuff_A()
        >>>     # Rollback changes to savepoint (the savepoint is still there)
        >>>     connection.savepoint_rollback(savepoint_id)
        >>>     # UoW collects events of the following function
        >>>     do_stuff_B()

    Despite the do_stuff_A changes were discarded, their events were emitted and
    handled by the bus. Although this scenario may be easy to cover, we are limiting
    flexibility on purpose to keep the code and uses simple:

        - Wrap all your atomic with a UoW context

    What happens with `atomic` blocks of third party applications? Well, since
    Django does not provide a rollback hook (e.g., signal) the only way to provide
    a solution is patching the Atomic class. We had a working implementation of
    this but we aren't sure this is the reasonable way to address the issue.

        - So, third party (rolled back) `atomic` blocks may cause issues.

    Other scenarios that may cause issues are:

        - Commiting or rolling back to a previous savepoint (not the last)
        - Using autocommit OFF
        - Closing and reopening the connection amidst an UoW transaction

    If you never thought about this issues before, then you don't need to worry
    about them. Code using autocommit OFF or rolling back to a concrete
    savepoint ID is likely an sysadmin script performing schema changes or
    doing something really, really, really special.

    We, as developers, usually think about a transaction as a context block.
    """

    using: list[str]

    def __init__(
        self,
        bus: "MessageBus",
        events_collector_cls: type["EventsCollector"] = DedupeEventsFifo,
        bind_connections: Using = None,
    ):
        """
        :param bind_connections: Bind the UoW/Bus to these database connections.
                                 Defaults to ["default"].
        """
        super().__init__(bus, events_collector_cls=events_collector_cls)

        if not bind_connections:
            self.using = [DEFAULT_DB_ALIAS]
        elif isinstance(bind_connections, str):
            self.using = [bind_connections]
        elif isinstance(bind_connections, list):
            self.using = bind_connections

    @contextmanager
    def atomic(self, using=None):
        with self:
            with atomic(using=using):
                yield

    def _begin(self):
        if not self.stack:
            self._bind_connections()
        super()._begin()

    def _end(self):
        transaction = super()._end()
        if not self.stack:
            self._unbind_connections()
        return transaction

    def _bind_connections(self):
        for using in self.using:
            connection = get_connection(using)
            connection.uow = self

    def _unbind_connections(self):
        for using in self.using:
            connection = get_connection(using)
            connection.uow = None



import inspect
import logging
import typing as t
from collections import defaultdict

from mb.commands import Command
from mb.events import Event
from mb.exceptions import DuplicatedCommandHandler
from mb.exceptions import InvalidMessage
from mb.exceptions import InvalidMessageType
from mb.exceptions import MissingCommandHandler
from mb.unit_of_work import UnitOfWork
from mb.injection import PreparedHandler

logger = logging.getLogger(__name__)


#
# NOTE: subscrbing to Event means subscrbing to all events. In the same way,
# subscribing to ParentEvent means subscribing to any ChildEvent(ParentEvent).
# I dunno if this makes sense for everyone, I'm just exploring options and
# trying in a real project to get a real experience.
# Btw, I'll better check/revise how this works in Kafka or Rabbit and get some
# clever ideas from the real pros.
#


class MessageBus:
    """
    An in-memory message bus.
    """

    event_handlers: dict[type[Event], list[t.Callable]]
    command_handlers: dict[type[Command], t.Callable]

    def __init__(self):
        self.event_handlers = defaultdict(list)
        self.command_handlers = {}

    def subscribe_event(self, cls: type[Event], handler: t.Callable) -> None:
        """
        Subscribe to an event type. An event may have multiple handlers

        :raises InvalidMessageType:
        """
        if not self.__is_event_type(cls):
            raise InvalidMessageType(f"This is not an event class: '{cls}'")

        if handler in self.event_handlers:
            logger.info(
                "Ignoring duplicated subscribe of handler '{handler}' to '{cls}'"
            )
        else:
            self.event_handlers[cls].append(handler)

    def subscribe_command(self, cls: type[Command], handler: t.Callable) -> None:
        """
        Set the command handler for this command. Only one handler allowed.

        :raises InvalidMessageType:
        :raises ConfigError: if there is already a handler
        """
        if not self.__is_command_type(cls):
            raise InvalidMessageType(f"This is not a command class: '{cls}'")

        current = self.command_handlers.get(cls, None)
        if current is None:
            self.command_handlers[cls] = handler
        elif current == handler:
            logger.info(
                "Ignoring duplicated subscribe of handler '{handler}' to '{cls}'"
            )
        else:
            raise DuplicatedCommandHandler(
                f"Duplicated handler for command '{cls}'. "
                f"The handler '{handler}' overrides the current '{current}'"
            )

    def __is_command_type(self, thing):
        return isinstance(thing, type) and issubclass(thing, Command)

    def __is_event_type(self, thing):
        return isinstance(thing, type) and issubclass(thing, Event)

    def _handle_command(self, command: Command, uow: UnitOfWork) -> t.Any:
        """
        Triggers the command handler. Handler exceptions are propagated

        :raises InvalidMessage: if this isn't a Command
        :raises MissingCommandHandler: if there's no handler configured for the command
        """
        if not isinstance(command, Command):
            raise InvalidMessage(f"This is not an command: '{command}'")

        try:
            handler = self.command_handlers[type(command)]
        except KeyError:
            raise MissingCommandHandler(command)

        logger.debug(f"Handling command '{command}': calling '{handler}'")
        prepared = PreparedHandler(handler, command, uow)
        return prepared()

    def _handle_event(self, event: Event, uow: UnitOfWork) -> None:
        """
        Triggers the event handlers. Handler exceptions are captured and error-logged.
        """
        # Collect handlers
        seen: set[t.Callable] = set()
        handlers: list[t.Callable] = []
        for event_cls in inspect.getmro(type(event)):
            for handler in self.event_handlers[event_cls]:
                if handler not in seen:
                    handlers.append(handler)
                    seen.add(handler)

        # Call handlers
        logger.debug(f"Handling event '{event}': {len(handlers)} handlers")
        for handler in handlers:
            logger.debug(f"Handling event '{event}': calling '{handler}'")
            prepared = PreparedHandler(handler, event, uow)
            try:
                prepared()
            except Exception:
                logger.exception(f"Exception handling event '{event}'")

    def clone(self) -> "MessageBus":
        """
        Returns a copy of itself with the same (detached) configuration
        """
        bus = type(self)()

        for command_cls, handler in self.command_handlers.items():
            bus.subscribe_command(command_cls, handler)

        for event_cls, handlers in self.event_handlers.items():
            for event_handler in handlers:
                bus.subscribe_event(event_cls, event_handler)
        return bus

    def _clear_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self._clear_command_handlers()
        self._clear_event_handlers()

    def _clear_command_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self.command_handlers = {}

    def _clear_event_handlers(self):
        """
        Clear handlers. Mainly for testing purposes.
        """
        self.event_handlers = defaultdict(list)



import typing as t

from django.db.models import Model
from django.db.transaction import get_connection

from mb.exceptions import UowContextRequired

if t.TYPE_CHECKING:
    from mb.unit_of_work.base import UnitOfWork


class BusModel(Model):
    class Meta:
        abstract = True

    @property
    def uow(self) -> "UnitOfWork":
        connection = get_connection(using=self._state.db)

        uow: t.Optional["UnitOfWork"] = getattr(connection, "uow", None)
        if not uow:
            raise UowContextRequired(
                "This instance is not attached ot any UnitOfWork. "
                "Did you forget opening the bus transaction (with uow:...)? "
            )
        return uow



import inspect
import typing as t

from mb.exceptions import CQProgrammingError, DuplicatedCommandHandler
from mb.exceptions import InjectError
from mb.messages import Message
from mb.unit_of_work import UnitOfWork

Args = tuple[t.Any, ...]
Kwargs = dict[str, t.Any]

P = t.ParamSpec("P")
R = t.TypeVar("R")


class PreparedHandler(t.Generic[P, R]):
    handler: t.Callable[P, R]
    arguments: inspect.BoundArguments

    def __init__(self, handler: t.Callable[P, R], message: Message, uow: UnitOfWork):
        self.handler = handler  # type: ignore
        self.arguments = self._prepare_arguments(handler, message, uow)

    def __call__(self) -> R:
        return self.handler(*self.args, **self.kwargs)

    @property
    def args(self) -> Args:
        return self.arguments.args

    @property
    def kwargs(self) -> Kwargs:
        return self.arguments.kwargs

    def _prepare_arguments(
        self,
        handler: t.Callable,
        message: Message,
        uow: UnitOfWork,
    ) -> inspect.BoundArguments:
        """
        Returns the arguments to inject into the handler
        """
        signature = inspect.signature(handler)

        not_defined = object()

        args: list[t.Any] = []
        kwargs: Kwargs = {}
        for name, param in signature.parameters.items():
            value = not_defined
            # try matching by annotation (MUST BE A REAL TYPE, NOT STR)
            if annotation := param.annotation:
                if isinstance(annotation, type):
                    if issubclass(annotation, Message):
                        value = message
                    elif issubclass(annotation, UnitOfWork):
                        value = uow
            # try matching by name
            if value is not_defined:
                if name in ("message", "command", "cmd", "event"):
                    value = message
                elif name in ("uow", "unit_of_work"):
                    value = uow
            # try default
            if value is not_defined:
                if param.default is not inspect.Parameter.empty:
                    value = param.default
            # can't do anything
            if value is not_defined:
                raise InjectError(name, handler)
            # collect
            if param.POSITIONAL_ONLY:
                args.append(value)
            else:
                kwargs[name] = value

        try:
            return signature.bind(*args, **kwargs)
        except TypeError:
            raise CQProgrammingError("Unexpected injection error")



import typing as t

from mb.commands import Command
from mb.events import Event
from mb.exceptions import InvalidMessage

if t.TYPE_CHECKING:
    from mb.bus import MessageBus


class UnitOfWork:
    """
    Simplest implementation. Handles events as they're emitted.
    """

    bus: "MessageBus"

    def __init__(self, bus: "MessageBus"):
        self.bus = bus

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def handle_command(self, command: Command) -> t.Any:
        """
        Triggers the command handler. Handler exceptions are propagated

        :raises InvalidMessage: if this isn't an Event
        :raises MissingCommandHandler: if there's no handler configured for the command
        """
        if not isinstance(command, Command):
            raise InvalidMessage(f"This is not a command: '{command}'")
        return self._handle_command(command)

    def _handle_command(self, command: Command) -> t.Any:
        return self.bus._handle_command(command, self)

    def emit_event(self, event: Event):
        """
        Triggers the event handlers. Handler exceptions are captured and error-logged.

        :raises InvalidMessage: if this isn't an Event
        """
        if not isinstance(event, Event):
            raise InvalidMessage(f"This is not an event: '{event}'")
        self._emit_event(event)

    def _emit_event(self, event: Event):
        self._handle_event(event)

    def _handle_event(self, event: Event):
        self.bus._handle_event(event, self)



import abc
import typing as t

if t.TYPE_CHECKING:
    from mb.events import Event


class EventsCollector(t.Collection["Event"], abc.ABC):
    """
    Collects events to be handled later
    """

    def __bool__(self):
        return bool(len(self))

    @abc.abstractmethod
    def push(self, event: "Event"):
        raise NotImplementedError()

    @abc.abstractmethod
    def pop(self) -> "Event":
        raise NotImplementedError()

    @abc.abstractmethod
    def extend(self, events: "EventsCollector") -> "Event":
        raise NotImplementedError()

    @abc.abstractmethod
    def clear(self):
        raise NotImplementedError()


class EventsFifo(EventsCollector):
    """
    FIFO
    """

    queue: list["Event"]

    def __init__(self):
        self.queue = []

    def push(self, event: "Event"):
        self.queue.append(event)

    def pop(self) -> "Event":
        return self.queue.pop()

    def extend(self, events: "EventsCollector") -> "Event":
        self.queue.extend(events)

    def clear(self):
        self.queue = []

    def __len__(self):
        return len(self.queue)

    def __iter__(self):
        return iter(self.queue)

    def __contains__(self, event: "Event"):  # type: ignore
        return event in self.queue


class DedupeEventsFifo(EventsCollector):
    """
    Deduplicated FIFO
    """

    queue: list["Event"]
    seen: set["Event"]

    def __init__(self):
        self.queue = []
        self.seen = set()

    def push(self, event: "Event"):
        if event not in self:
            self.seen.add(event)
            self.queue.append(event)

    def pop(self) -> "Event":
        event = self.queue.pop()
        self.seen.remove(event)
        return event

    def extend(self, events: "EventsCollector") -> "Event":
        for event in events:
            self.push(event)

    def clear(self):
        self.queue = []
        self.seen = set()

    def __len__(self):
        return len(self.queue)

    def __iter__(self):
        return iter(self.queue)

    def __contains__(self, event: "Event"):  # type: ignore
        return event in self.seen



import typing as t

from mb.exceptions import DuplicatedMessageName
from mb.exceptions import InvalidMessageName


class MessageMeta(type):
    """
    The Message metaclass.

    Message.NAME defaults to __module__.__name__
    Message.NAME type and collision check
    """

    _messages: dict[str, "MessageMeta"] = {}

    def __new__(cls, name, bases, dic, **kw):
        cls._set_default_name(name, dic)
        cls._check_name_type(name, dic)
        cls._check_unique_name(name, dic)

        # Create & register
        message_cls = super().__new__(cls, name, bases, dic, **kw)
        cls._messages[message_cls.NAME] = message_cls
        return message_cls

    @classmethod
    def _set_default_name(cls, name, dic):
        try:
            dic["NAME"]
        except KeyError:
            dic["NAME"] = f"{dic['__module__']}.{name}"

    @classmethod
    def _check_name_type(cls, name, dic):
        message_name = dic["NAME"]
        if not isinstance(message_name, str):
            raise InvalidMessageName(
                f"'{name}.NAME' must be of 'str' type. Found: '{type(message_name)}'"
            )

    @classmethod
    def _check_unique_name(cls, name, dic):
        message_name = dic["NAME"]
        try:
            message_cls = cls._messages[message_name]
        except KeyError:
            return
        raise DuplicatedMessageName(
            "These messages have the same NAME: "
            f"'{name}' and '{message_cls.__name__}'"
        )

    @classmethod
    def _clear(cls):
        """
        Resets internal state. For testing.
        """
        cls._messages = {}


class Message(metaclass=MessageMeta):
    """
    The Message base class
    """

    NAME: t.ClassVar[str]



from mb.contrib.django.unit_of_work import DjangoUnitOfWork
from mb.unit_of_work import UnitOfWork
from app.common.bus.bus import bus


class UnitOfWorkInjectionMiddleware:
    """
    Bootstaps and injects a new UnitOfWork in the request.META
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        # Tests may inject its own UnitOfWork
        uow = request.META.get("uow")
        if not isinstance(uow, UnitOfWork):
            request.META["uow"] = DjangoUnitOfWork(bus)

        return self.get_response(request)
